/* ==========================================================================
 *
 *                         ---=MOD_SAMOYLYK=---
 *                            Version 1.0.2
 *
 * Copyright (C) 2003, 2004  Oleksandr Samoylyk (oleksandr@samoylyk.sumy.ua)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * ==========================================================================
 */

// Header's files
#include "httpd.h"
#include "http_config.h"
#include "http_request.h"
#include "http_core.h"
#include "http_log.h"
#include "http_main.h"
#include "http_protocol.h"
#include "util_script.h"
#include "http_conf_globals.h"

// Module version
#define MODULE_VERSION "1.0.2"

MODULE_VAR_EXPORT module samoylyk_module;

typedef struct samoylyk_ll_node samoylyk_ll_node;
typedef struct samoylyk_ll_head samoylyk_ll_head;
typedef struct samoylyk_alias samoylyk_alias;

// Alias linked list
struct samoylyk_alias
{
        char* is;
        char* should;
        samoylyk_alias* next;
};

// Virtual Host node
struct samoylyk_ll_node
{
        samoylyk_ll_node* next;
        samoylyk_ll_head* head;

        char* hostName;
        char* user;
        char* group;
        char* root;
        samoylyk_alias* aliasList;

};

// Virtual Host linked list
struct samoylyk_ll_head
{
        int numNodes;

        samoylyk_alias* gAliasList;
        samoylyk_ll_node* start;
        samoylyk_ll_node* end;
};

// Configuration structure
typedef struct
{
        samoylyk_ll_head* head;
        pool* master_pool;
        pool* samoylyk_pool;
        char * conf_file;
        char * base_path;
        char * log_file;
        long modified;
} samoylyk_config;

///////////////////////////////////////////////////////////
//////////////////////Internal Functions///////////////////
///////////////////////////////////////////////////////////

// Take off the slash at the end (read the INSTALL's notes)

int Samoylyk_TrimSlash(char* in)
{
        int i = strlen(in);

        if( in[i-1] == '/')
                in[i-1] = 0;

return(0);
}

// Log
int Samoylyk_Log(samoylyk_config* cfg, const char* error,...)
{

        if (cfg->log_file) {
                va_list args;
                char com[2048];
                FILE* fout;

                va_start(args, error);
                vsprintf(com, error, args);

                if( (fout = fopen(cfg->log_file,"a")) == NULL)
                        printf("Error opening log file\n");
                else {
                        fprintf(fout,"%s\n",com);
                }
                fclose(fout);
                va_end(args);
        }

return(1);
}

// Initializes a Virtual Host node
samoylyk_ll_node* Samoylyk_CreateNode(samoylyk_config* cfg)
{
        samoylyk_ll_node* ret = (samoylyk_ll_node*)ap_pcalloc(cfg->samoylyk_pool, sizeof(samoylyk_ll_node));
        ret->aliasList = (samoylyk_alias*) ap_pcalloc(cfg->samoylyk_pool, sizeof(samoylyk_alias));
        ret->head = cfg->head;

return(ret);
}

// Initializes the Virtual Host linked list head node
samoylyk_ll_head* Samoylyk_CreateHeadNode(samoylyk_config* cfg)
{
        samoylyk_ll_head* ret = (samoylyk_ll_head*)ap_pcalloc(cfg->samoylyk_pool, sizeof(samoylyk_ll_node));

return(ret);
}

//Initializes the Virtual Host linked list
int Samoylyk_InitLL(samoylyk_config* cfg)
{
        cfg->samoylyk_pool = ap_make_sub_pool(cfg->master_pool);
        cfg->head = Samoylyk_CreateHeadNode(cfg);
        cfg->head->start = Samoylyk_CreateNode(cfg);
        cfg->head->end = cfg->head->start;
        cfg->head->numNodes = 0;
        cfg->head->gAliasList = (samoylyk_alias*) ap_pcalloc(cfg->samoylyk_pool, sizeof(samoylyk_alias));

return(0);
}

char* Samoylyk_ParseTag(char* line, char* req)
{
        char* ret = line;

        while(*req)
                if(*ret++ != *req++)
                        return(NULL);
	
return(ret+1);
}

// Add an Alias to the alias list
int Samoylyk_AddAlias(samoylyk_alias* dest, char* pIs, char* pShould, pool* p)
{
        if(dest->next)
        {
                Samoylyk_AddAlias(dest->next, pIs, pShould, p);
        }
        else
        {
                Samoylyk_TrimSlash(pIs);
                dest->is = pIs;
                dest->should = pShould;
                dest->next = (samoylyk_alias*) ap_pcalloc(p, sizeof(samoylyk_alias));
        }

return(0);
}

// Config file
int Samoylyk_LoadConfigFile(samoylyk_config* cfg)
{
        FILE* fin;
        char line[2048];
        char* param;
        char tmp1[2048];
        char tmp2[2048];
        struct stat buf;
        long timer;
        Samoylyk_Log(cfg, "Loading in config file...");

        Samoylyk_InitLL(cfg);

        if (stat(cfg->conf_file,&buf))
        return(Samoylyk_Log(cfg, "Config file does not exist"));

        cfg->modified = buf.st_mtime;

    timer = time(NULL);
    while ((fin = fopen(cfg->conf_file, "r")) == NULL && timer + 20 > time(NULL)); 
    if(fin == NULL)
        return(Samoylyk_Log(cfg, "Error opening config file."));

        while(!feof(fin))
        {
                memset(line,0,2048);
                if(fgets(line, 2048,fin) == NULL)
                        if(!feof(fin))
                                return(Samoylyk_Log(cfg, "Error Reading in line from config file."));

                if(line[strlen(line)-1] == '\n')
                        line[strlen(line)-1] = 0;
                if(line[0] == '#') {
                }
                else if((param = Samoylyk_ParseTag(line,"GAlias")))
                {
                        sscanf(param,"%s %s",tmp1, tmp2);
                        Samoylyk_AddAlias(cfg->head->gAliasList, ap_pstrdup(cfg->samoylyk_pool,tmp1), ap_pstrdup(cfg->samoylyk_pool,tmp2), cfg->samoylyk_pool);
                }
                else if((param = Samoylyk_ParseTag(line,"ServerName")))
                {
                        cfg->head->end->hostName = ap_pstrdup(cfg->samoylyk_pool, param);
                }
                else if((param = Samoylyk_ParseTag(line,"User")))
                {
                        cfg->head->end->user = ap_pstrdup(cfg->samoylyk_pool, param);
                }
                else if((param = Samoylyk_ParseTag(line,"Group")))
                {
                        cfg->head->end->group = ap_pstrdup(cfg->samoylyk_pool, param);
                }
                else if((param = Samoylyk_ParseTag(line,"DocumentRoot")))
                {
                        cfg->head->end->root = ap_pstrdup(cfg->samoylyk_pool, param);
                }
                else if((param = Samoylyk_ParseTag(line,"Alias")))
                {
                        sscanf(param,"%s %s",tmp1, tmp2);
                        Samoylyk_AddAlias(cfg->head->end->aliasList, ap_pstrdup(cfg->samoylyk_pool,tmp1), ap_pstrdup(cfg->samoylyk_pool,tmp2), cfg->samoylyk_pool);
                }
                else if((param = Samoylyk_ParseTag(line,"End")))
                {
                        cfg->head->end->next = Samoylyk_CreateNode(cfg);
                        cfg->head->end = cfg->head->end->next;
                        cfg->head->numNodes++;
                }
                else
                {
                        if(*line)
                                Samoylyk_Log(cfg, "Unknown Directive found in config file.");
                }
	}

        Samoylyk_Log(cfg, "Finished, total of %d virtual servers loaded.", cfg->head->numNodes);
        
        fclose(fin);
return(0);
}

int Samoylyk_UnLoadConfigFile(samoylyk_config* cfg)
{
        Samoylyk_Log(cfg, "Unloading Config file...");

        if(cfg->samoylyk_pool)
                ap_destroy_pool(cfg->samoylyk_pool);

        cfg->modified = 0;
        

        Samoylyk_Log(cfg, "Finished unloading config file.");

return(0);
}

samoylyk_ll_node* Samoylyk_GetNode(samoylyk_config* cfg, const char* host, samoylyk_ll_node* pNode)
{
	if(strcmp(pNode->hostName, host) == 0)
		return(pNode);
	else {
		if(pNode->next->next == NULL)
		{
			while(*host++ != '.')
				if(*host == 0)
					return(NULL);
			
			return(Samoylyk_GetNode(cfg, host, pNode->head->start));
		}	
		else
			return(Samoylyk_GetNode(cfg, host, pNode->next));
	}
return(NULL);
}

int Samoylyk_URIMatch(char* alias, char* uri)
{
	int lenAlias = strlen(alias);
	int lenUri = strlen(uri);
	int i;

	if(lenAlias > lenUri)
		return(0);

	for(i=0; i < lenAlias; i++)
	{
		if( alias[i] != uri[i])
			return(0);
	}

	if( (uri[i] == 0) || (uri[i] == '/') )
		return(1);

return(0);
}

samoylyk_alias* Samoylyk_IsAlias(samoylyk_alias* n, char* uri)
{
	if((!*uri)||(n->next==NULL))
		return(NULL);

	if( Samoylyk_URIMatch( n->is, uri))
		return(n);
	else
		if(n->next->next == NULL)
			return(NULL);	
		else
			return(Samoylyk_IsAlias(n->next, uri));

return(NULL);
}

char* Samoylyk_TrimURI(char* p, char* uri)
{
	while(*p)
	{
		p++;
		uri++;
	}
	
return(uri);
}

int Samoylyk_SetGroupID(char* arg, request_rec* cmd)
{

    cmd->server->server_gid = ap_gname2id(arg);

return(0);
}

int Samoylyk_SetUserID(char* arg, request_rec* cmd)
{
	ap_user_name =  ap_pstrdup(cmd->pool, arg);

	cmd->server->server_uid = ap_uname2id(arg);

return(0);
}

///////////////////////////////////////////////////////////////////////
//////////////////////Functions that Apache Calls//////////////////////
///////////////////////////////////////////////////////////////////////

static void init_module(server_rec *s, pool *p)
{
    ap_add_version_component("mod_samoylyk/" MODULE_VERSION);
}

static const char *cmd_samoylykconf(cmd_parms *cmd, void *dconf, char *str)
{
    samoylyk_config* cfg;

    cfg = (samoylyk_config*)
            ap_get_module_config(cmd->server->module_config, &samoylyk_module);

    cfg->conf_file = ap_pstrdup(cfg->master_pool, str);

    return NULL;
}

static const char *cmd_samoylykbase(cmd_parms *cmd, void *dconf, char *str)
{
    samoylyk_config* cfg;

    cfg = (samoylyk_config*)
            ap_get_module_config(cmd->server->module_config, &samoylyk_module);

	Samoylyk_TrimSlash(str);
    cfg->base_path = ap_pstrdup(cfg->master_pool, str);

    return NULL;
}

static const char *cmd_samoylyklog(cmd_parms *cmd, void *dconf, char *str)
{
    samoylyk_config* cfg;

    cfg = (samoylyk_config*)
            ap_get_module_config(cmd->server->module_config, &samoylyk_module);

    cfg->log_file = ap_pstrdup(cfg->master_pool,str);

    return NULL;
}

static const command_rec command_table[] = {
    { "SAMOYLYKConfFile",     cmd_samoylykconf,     NULL, RSRC_CONF,   TAKE1,
      "Path to the samoylyk dynamic virtual-host database" },
    { "SAMOYLYKBasePath",     cmd_samoylykbase,     NULL, RSRC_CONF,   TAKE1,
      "Path to the base document root" },
    { "SAMOYLYKLogFile",      cmd_samoylyklog,      NULL, RSRC_CONF,   TAKE1,
      "Log file path" },
    { NULL }
};

void* ServerConfig(pool* p, server_rec* s)
{	
	samoylyk_config* config = (samoylyk_config*) ap_pcalloc(p, sizeof(samoylyk_config) );
	config->master_pool = p;
	config->base_path = ap_pstrdup(config->master_pool, "");
	return((void*)config);
}


int FilenameTranslation(request_rec *r)
{
	char* res;
	char* uri;
	struct stat buf;
	samoylyk_ll_node* reqNode;
	samoylyk_alias* a;

	samoylyk_config* cfg = (samoylyk_config*)ap_get_module_config (r->server->module_config,
					       &samoylyk_module);
	
	if(!cfg)
		return(DECLINED);

	Samoylyk_Log(cfg, "-------------------------------------");
	Samoylyk_Log(cfg, "Processing request: %s",r->the_request);
	Samoylyk_Log(cfg, "HostName: %s", r->hostname);
	Samoylyk_Log(cfg, "URI: %s", r->uri);
	Samoylyk_Log(cfg, "Server Hostname: %s",r->server->server_hostname);

	if (stat(cfg->conf_file,&buf)) {
		Samoylyk_Log(cfg, "Config file does not exist");
		return(DECLINED);
	}
	if(cfg->modified != buf.st_mtime)
	{
		Samoylyk_Log(cfg, "Changes detected on config file, re-loading.");
		Samoylyk_UnLoadConfigFile(cfg);
		Samoylyk_LoadConfigFile(cfg);
	}
	reqNode = Samoylyk_GetNode(cfg, r->hostname, cfg->head->start);
	if(!reqNode)
	{
		Samoylyk_Log(cfg, "Couldn't locate virtual server node, unknown host");
		return(HTTP_NOT_FOUND);
	}

	uri = ap_pstrdup(r->pool, r->uri);
	Samoylyk_TrimSlash(uri);
	if( (a = Samoylyk_IsAlias(reqNode->head->gAliasList, uri)) )
	{
		res = ap_pstrdup(r->pool, a->should);
		Samoylyk_Log(cfg, "Request is a Global Alias, after translation : %s",res);
		Samoylyk_TrimSlash(res);
		uri = Samoylyk_TrimURI(a->is, uri);
	}
	else if( (a = Samoylyk_IsAlias(reqNode->aliasList, uri)) )
	{
		res = ap_pstrdup(r->pool, a->should);		
		Samoylyk_Log(cfg, "Request is an Alias, after translation : %s",res);
		Samoylyk_TrimSlash(res);
		uri = Samoylyk_TrimURI(a->is, uri);
	}
	else
	{
		res = ap_pstrdup(r->pool, reqNode->root);
		Samoylyk_TrimSlash(res);
	}

	res = ap_pstrcat(r->pool, res, uri, NULL);
	r->filename = ap_pstrcat(r->pool, cfg->base_path, res, NULL);

	r->server->server_hostname = ap_pstrdup(r->pool, reqNode->hostName);

	Samoylyk_SetGroupID( reqNode->group, r);
	Samoylyk_SetUserID( reqNode->user, r);

	Samoylyk_Log(cfg, "User: %d", r->server->server_uid);
	Samoylyk_Log(cfg, "Group: %d", r->server->server_gid);
	Samoylyk_Log(cfg, "Node Root + URI: %s", res);
	Samoylyk_Log(cfg, "Final Path: %s", ap_pstrcat(r->pool, cfg->base_path, res, NULL));

	return(OK);
}

/*
Apache exports information
*/
module MODULE_VAR_EXPORT samoylyk_module =
{
    STANDARD_MODULE_STUFF,
    init_module,                                /* module initializer */
    NULL,                                       /* per-directory config creator */
    NULL,                                       /* dir config merger */
    ServerConfig,                               /* server config creator */
    NULL,                                       /* server config merger */
    command_table,                              /* command table */
    NULL,                                       /* list of handlers */
    FilenameTranslation,                        /* URI-to-filename translation */
    NULL,                                       /* check/validate user_id */
    NULL,                                       /* check auth */
    NULL,                                       /* check access by host address */
    NULL,                                       /* type_checker */
    NULL,                                       /* fixups */
    NULL,                                       /* logger */
    NULL,                                       /* header parser */
    NULL,                                       /* child_init */
    NULL,                                       /* child_exit */
    NULL                                        /* post read-request */
};
